/*
xmlFoam - an OpenFOAM implementation of a proposed XML-based geometry /
condition description format for CFD

Copyright (C) 2010  Takuya OSHIMA <oshima@eng.niigata-u.ac.jp>.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301, USA.

OPENFOAM(R) is a registered trade mark of OpenCFD Limited, the
producer of the OpenFOAM software and owner of the OPENFOAM(R) and
OpenCFD(R) trade marks. This offering is not approved or endorsed by
OpenCFD Limited.
*/

#include "xmlBoundaries.h"
#include "xmlCFD.h"

#include "GeometricFields.H"
#include "Time.H"
#include "argList.H"
#include "polyMesh.H"
#include "timeSelector.H"
#include "volFields.H"

#include "vtkCell.h"
#include "vtkCellData.h"
#include "vtkFloatArray.h"
#include "vtkPolyData.h"

using namespace Foam;

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

// Find face on patch. Return mesh face label.
label findFace
(
    const primitivePatch& pp,
    const point& nearPoint
)
{
    const pointField& points = pp.points();

    // Find nearest and next nearest
    scalar minDist = GREAT;
    label minI = -1;

    scalar almostMinDist = GREAT;
    label almostMinI = -1;

    forAll(pp, patchFaceI)
    {
        pointHit pHit(pp[patchFaceI].nearestPoint(nearPoint, points));

        if (pHit.hit())
        {
            if (pHit.distance() < minDist)
            {
                almostMinDist = minDist;
                almostMinI = minI;

                minDist = pHit.distance();
                minI = patchFaceI;
            }
            else if (pHit.distance() < almostMinDist)
            {
                almostMinDist = pHit.distance();
                almostMinI = patchFaceI;
            }
        }
    }

    if (minI == -1)
    {
        Info<< "Did not find face close to point " << nearPoint << endl;

        return -1;
    }

    // Decide if nearPoint unique enough.
    Info<< "Found to point " << nearPoint << nl
        << "    nearest face      : " << minI
        << " distance " << minDist
        << " to face centre " << pp[minI].centre(points) << nl
        << "    next nearest face : " << almostMinI
        << " distance " << almostMinDist
        << " to face centre " << pp[almostMinI].centre(points) << nl
        << endl;

    if (almostMinDist < 2*minDist)
    {
        Info<< "Next nearest too close to nearest. Aborting" << endl;

        return -1;
    }
    else
    {
        return minI;
    }
}

template<int nCmpts, typename Type>
inline void substitute(fvPatchField<Type> &pf, const label faceI,
    vtkFloatArray *data, const vtkIdType cellId);

template<>
inline void substitute<1, scalar>(fvPatchField<scalar> &pf, const label faceI,
vtkFloatArray *data, const vtkIdType cellId)
{
    pf[faceI] = data->GetValue(cellId);
}

template<>
inline void substitute<3, vector>(fvPatchField<vector> &pf, const label faceI,
vtkFloatArray *data, const vtkIdType cellId)
{
    float *p = data->GetPointer(3 * cellId);
    pf[faceI].x() = p[0];
    pf[faceI].y() = p[1];
    pf[faceI].z() = p[2];
}

template<int nCmpts, typename Type>
inline void substitute(fvPatchField<Type> &pf, const label faceI,
    vtkFloatArray *data, const vtkIdType cellId)
{
    FatalErrorIn("substitute()") << "nCmpts = " << nCmpts << ": Not implemented"
        << exit(FatalError);
}

template<int nCmpts, typename Type>
GeometricField<Type, fvPatchField, volMesh> *
mapFieldData(xmlBoundary &boundary, const label fieldI, const vector &tl,
    const label patchID, const fvMesh &mesh, const Time &runTime)
{
    xmlField &field = boundary[fieldI];
    vtkstd::vector<int> &dim = field.Dimensions;
    vtkPolyData *fieldData = boundary.FieldData;
    const primitivePatch &pp = mesh.boundaryMesh()[patchID];

    GeometricField<Type, fvPatchField, volMesh> *vfPtr
        = new GeometricField<Type, fvPatchField, volMesh>
        (
            IOobject
            (
                field.Identifier,
                runTime.timeName(),
                mesh,
                IOobject::READ_IF_PRESENT,
                IOobject::AUTO_WRITE
            ),
            mesh,
            dimensioned<Type>(field.Identifier, dimensionSet(dim[0], dim[1],
            dim[2], dim[3], dim[4], dim[5], dim[6]), pTraits<Type>::zero)
        );
    GeometricField<Type, fvPatchField, volMesh> &vf = *vfPtr;

    vf.boundaryField().set
    (
        patchID,
        fvPatchField<Type>::New
        (
            field.Conditions,
            mesh.boundary()[patchID],
            vf
	)
    );
    fvPatchField<Type> &pf = vf.boundaryField()[patchID];

    if(field.UseFieldData)
    {
        vtkFloatArray *data = vtkFloatArray::SafeDownCast(
            fieldData->GetCellData()->GetArray(field.FieldNameInData.c_str()));
        if(!data)
        {
            FatalErrorIn("main()") << "Field " << field.FieldNameInData
                << " not found in field data" << exit(FatalError);
        }
#if 0
        const label nFaces
            = static_cast<label>(fieldData->GetNumberOfCells());
        for(label faceI = 0; faceI < nFaces; ++faceI)
        {
            // Get face centroid
            vtkCell *f = fieldData->GetCell(faceI);
            double pCenter[3], centroid[3], weights[128];
            int subId;
            subId = f->GetParametricCenter(pCenter);
            f->EvaluateLocation(subId, pCenter, centroid, weights);
            const point p(vector(centroid[0], centroid[1], centroid[2]) + tl);
            const label matchingFaceI = findFace(pp, p);

            if(matchingFaceI == -1)
            {
                FatalErrorIn("mapFieldData()") << "Could not find matching"
                    " face in OpenFOAM mesh for face " << faceI
                    << " in field data" << exit(FatalError);
            }
            substitute<nCmpts, Type>(pf, matchingFaceI, data, faceI);
        }
#else
        forAll(pp, faceI)
        {
            const point centroid(pp[faceI].centre(pp.points()));
            double p[3], pcoords[3], weights[128];
            p[0] = centroid.x() - tl.x();
            p[1] = centroid.y() - tl.y();
            p[2] = centroid.z() - tl.z();
            int subId;
            const vtkIdType cellId = fieldData->FindCell(p, NULL, 0, 0.001,
                subId, pcoords, weights);
            if(cellId < 0)
            {
                FatalErrorIn("mapFieldData()") << "Could not find matching"
                    " face in field data for face " << faceI
                    << " in OpenFOAM mesh" << exit(FatalError);
            }
            substitute<nCmpts, Type>(pf, faceI, data, cellId);
        }
#endif
    }

    return vfPtr;
}

int main(int argc, char *argv[])
{
    timeSelector::addOptions();
    argList::noParallel();
    argList::validArgs.append(".xml file");

#include "setRootCase.H"
#include "createTime.H"

    instantList timeDirs = timeSelector::select0(runTime, args);
    if(timeDirs.size() != 1)
    {
        FatalErrorIn("main()") << "Requires option -time <time>"
            << exit(FatalError);
    }
    runTime.setTime(timeDirs[0], 0);
    runTime.functionObjects().off();

    const fileName xmlFileName(args.additionalArgs()[0]);

    xmlCFD *parser = xmlCFD::New();
    parser->SetFileName(xmlFileName.c_str());

    if(!parser->Parse() || parser->GetBoundaries().size() == 0)
    {
        FatalErrorIn("main()") << "Failed to read the main XML file"
            << exit(FatalError);
    }

    xmlBoundaries &boundaries = parser->GetBoundaries();
    xmlBoundary &boundary = boundaries[0];

#include "createMesh.H"
#include "readXmlFoamDict.H"

    const label patchID = mesh.boundaryMesh().findPatchID(boundary.Identifier);
    if(patchID == -1)
    {
        FatalErrorIn("main()") << "Could not find matching patch in OpenFOAM"
            " mesh for patch " << boundary.Identifier << exit(FatalError);
    }

    const vector tl(xmlFoamDict.lookup("translationVector"));

    List<regIOobject*> fields(boundary.size());

    for(label fieldI = 0; fieldI < boundary.size(); ++fieldI)
    {
        xmlField &field = boundary[fieldI];

        vtkstd::vector<int> &dim = field.Dimensions;

        if(field.FieldType == xmlField::SCALAR)
        {
            fields[fieldI] = mapFieldData<1, scalar>(boundary, fieldI, tl,
                patchID, mesh, runTime);
        }
        else if(field.FieldType == xmlField::VECTOR)
        {
            fields[fieldI] = mapFieldData<3, vector>(boundary, fieldI, tl,
                patchID, mesh, runTime);
        }
    }
    parser->Delete();

    runTime++;
    runTime.writeNow();

    for(label fieldI = 0; fieldI < boundary.size(); ++fieldI)
    {
        delete fields[fieldI];
    }

    return 0;
}
