/*=========================================================================

  xmlCFD - an implementation of a proposed XML-based geometry /
  condition description format for CFD

  Copyright (c) 2010 Takuya OSHIMA <oshima@eng.niigata-u.ac.jp>.
  All rights reserved.
  See License.txt for details.

     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notice for more information.

=========================================================================*/

#include "xmlMainWindow.h"
#include "ui_xmlMainWindow.h"

#include "xmlBoundaries.h"
#include "xmlCFD.h"

#include "vtkActor.h"
#include "vtkActorCollection.h"
#include "vtkCamera.h"
#include "vtkCellData.h"
#include "vtkInteractorStyleTrackballCamera.h"
#include "vtkPolyData.h"
#include "vtkPolyDataMapper.h"
#include "vtkRenderer.h"
#include "vtkRendererCollection.h"
#include "vtkRenderWindow.h"
#include "vtkRenderWindowInteractor.h"
#include "vtkSmartPointer.h"

#include <QtGui>

//------------------------------------------------------------------------------
xmlMainWindow::xmlMainWindow(QWidget *parent)
  : QMainWindow(parent), Ui(new Ui_xmlMainWindow), Parser(0)
{
  this->Ui->setupUi(this);

  // File -> Open menu item
  QObject::connect(this->Ui->actionOpen, SIGNAL(triggered()), this,
      SLOT(slotOpen()));

  // File -> Close menu item
  QObject::connect(this->Ui->actionClose, SIGNAL(triggered()), this,
      SLOT(slotClose()));

  // File -> Exit menu item
  QObject::connect(this->Ui->actionExit, SIGNAL(triggered()), this,
      SLOT(slotExit()));

  this->setupRenderWindow(this->Ui->qvtkGeometry);
  this->setupRenderWindow(this->Ui->qvtkModelledGeometry);
  this->setupRenderWindow(this->Ui->qvtkBoundaryGeometry);
  this->setupRenderWindow(this->Ui->qvtkFieldData);
}

//------------------------------------------------------------------------------
xmlMainWindow::~xmlMainWindow()
{
  if(this->Parser)
    {
    this->Parser->Delete();
    }
}

//------------------------------------------------------------------------------
void xmlMainWindow::setupRenderWindow(QVTKWidget *qvtk)
{
  // Create renderer
  vtkSmartPointer<vtkPolyDataMapper> mapper
    = vtkSmartPointer<vtkPolyDataMapper>::New();
  mapper->SetInput(0);

  vtkSmartPointer<vtkActor> actor = vtkSmartPointer<vtkActor>::New();
  actor->SetMapper(mapper);

  vtkSmartPointer<vtkRenderer> renderer = vtkSmartPointer<vtkRenderer>::New();
  renderer->AddActor(actor);
  renderer->SetBackground(0.1, 0.2, 0.4);

  qvtk->GetRenderWindow()->AddRenderer(renderer);
}

//------------------------------------------------------------------------------
void xmlMainWindow::setInput(QVTKWidget *qvtk, vtkPolyData *data)
{
    vtkRendererCollection *renderers = qvtk->GetRenderWindow()->GetRenderers();
    renderers->InitTraversal();
    for(vtkRenderer *renderer = renderers->GetNextItem(); renderer;
        renderer = renderers->GetNextItem())
      {
      vtkActorCollection *actors = renderer->GetActors();
      actors->InitTraversal();
      for(vtkActor *actor = actors->GetNextItem(); actor;
          actor = actors->GetNextItem())
        {
        vtkPolyDataMapper::SafeDownCast(actor->GetMapper())->SetInput(data);
        }
      renderer->ResetCamera();
      }
    qvtk->GetRenderWindow()->Render();
}

//------------------------------------------------------------------------------
void xmlMainWindow::createActions()
{
  // Field selection combo box
  // Add array names to the combo box
  vtkPolyData *fieldData
      = this->Parser->GetBoundaries()[0].FieldData;
  if(!fieldData)
    {
    return;
    }

  vtkCellData *arrays = fieldData->GetCellData();
  const int nArrays = arrays->GetNumberOfArrays();
  for(int arrayI = 0; arrayI < nArrays; ++arrayI)
    {
    this->Ui->comboFieldSelection->addItem(arrays->GetArray(arrayI)->GetName());
    }
  // Set the current item of the combo box to the active scalar
  int activeAttributeIndices[vtkDataSetAttributes::NUM_ATTRIBUTES];
  arrays->GetAttributeIndices(activeAttributeIndices);
  if(activeAttributeIndices[vtkDataSetAttributes::SCALARS] >= 0)
    {
    this->Ui->comboFieldSelection->setCurrentIndex(
        activeAttributeIndices[vtkDataSetAttributes::SCALARS]);
    }
  // Connect selection change to active scalar setting
  QObject::connect(this->Ui->comboFieldSelection,
      SIGNAL(currentIndexChanged(int)), this, SLOT(slotFieldSelection(int)));
}

//------------------------------------------------------------------------------
void xmlMainWindow::slotOpen()
{
  QString fileName = QFileDialog::getOpenFileName(this, tr("Open File"),
      QDir::currentPath(), tr("XML files (*.xml)"));

  if(fileName.isNull())
    {
    return;
    }

  // If we have a file already opened, close it first
  this->slotClose();

  this->Parser = xmlCFD::New();
  this->Parser->SetFileName(fileName.toLocal8Bit().data());
  if(!this->Parser->Parse() || this->Parser->GetBoundaries().size() == 0)
    {
    return;
    }

  this->setInput(this->Ui->qvtkGeometry,
      this->Parser->GetBoundaries().Geometry);
  this->setInput(this->Ui->qvtkModelledGeometry,
      this->Parser->GetBoundaries().ModelledGeometry);
  this->setInput(this->Ui->qvtkBoundaryGeometry,
      this->Parser->GetBoundaries()[0].BoundaryGeometry);
  this->setInput(this->Ui->qvtkFieldData,
      this->Parser->GetBoundaries()[0].FieldData);

  this->createActions();
}

//------------------------------------------------------------------------------
void xmlMainWindow::slotClose()
{
  if(this->Parser)
    {
    // Disconnect selection change to active scalar setting first,
    // otherwise currentIndexChanged(-1) will be sent to the slot.
    QObject::disconnect(this->Ui->comboFieldSelection, 0, this, 0);
    this->Ui->comboFieldSelection->clear();

    this->setInput(this->Ui->qvtkGeometry, 0);
    this->setInput(this->Ui->qvtkModelledGeometry, 0);
    this->setInput(this->Ui->qvtkBoundaryGeometry, 0);
    this->setInput(this->Ui->qvtkFieldData, 0);

    this->Parser->Delete();
    this->Parser = 0;
    }
}

//------------------------------------------------------------------------------
void xmlMainWindow::slotExit()
{
  qApp->exit();
}

//------------------------------------------------------------------------------
void xmlMainWindow::slotFieldSelection(int index)
{
  vtkCellData *arrays
      = this->Parser->GetBoundaries()[0].FieldData->GetCellData();
  if(arrays->GetArray(index)->GetNumberOfComponents() == 1)
    {
    arrays->SetActiveAttribute(index, vtkDataSetAttributes::SCALARS);
    }
  this->Ui->qvtkFieldData->GetRenderWindow()->Render();
}
