/*=========================================================================

  xmlCFD - an implementation of a proposed XML-based geometry /
  condition description format for CFD

  Copyright (c) 2010 Takuya OSHIMA <oshima@eng.niigata-u.ac.jp>.
  All rights reserved.
  See License.txt for details.

     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notice for more information.

=========================================================================*/

#include "xmlCFD.h"
#include "xmlBoundaries.h"

#include "vtkExecutive.h"
#include "vtkFloatArray.h"
#include "vtkObjectFactory.h"
#include "vtkPolyData.h"
#include "vtkStdString.h"
#include "vtkXMLPolyDataReader.h"

#include <vtksys/SystemTools.hxx>
#include <vtksys/ios/sstream>

vtkCxxRevisionMacro(xmlCFD, "$Revision: 1.00 $");
vtkStandardNewMacro(xmlCFD);

//------------------------------------------------------------------------------
xmlCFD::xmlCFD()
  : Boundaries(new xmlBoundaries), Path(new vtkStdString)
{
}

//------------------------------------------------------------------------------
xmlCFD::~xmlCFD()
{
  delete this->Boundaries;
  delete this->Path;
}

//------------------------------------------------------------------------------
void xmlCFD::PrintSelf(ostream& os, vtkIndent indent)
{
  this->Superclass::PrintSelf(os, indent);
}

//------------------------------------------------------------------------------
void xmlCFD::SetFileName(const char *fileName)
{
  this->Superclass::SetFileName(fileName);
  *this->Path = vtksys::SystemTools::CollapseFullPath(
      vtksys::SystemTools::GetFilenamePath(fileName).c_str());
}

//------------------------------------------------------------------------------
int xmlCFD::ParseName(vtkXMLDataElement *element)
{
  // Name: mandatory, can appear more than once
  const int nNestedElements = element->GetNumberOfNestedElements();
  bool nameFound = false;
  for(int elemI = 0; elemI < nNestedElements; ++elemI)
    {
    vtkXMLDataElement *elem = element->GetNestedElement(elemI);
    if(vtkStdString(elem->GetName()) == "Name")
      {
      nameFound = true;
      // lang: mandatory
      if(const char *lang = elem->GetAttribute("lang"))
        {
        vtkstd::cout << "Name (" << lang << "): " << elem->GetCharacterData()
                     << vtkstd::endl;
        }
      else
        {
        vtkstd::cout << "Mandatory attribute lang not found" << vtkstd::endl;
        return 0;
        }
      }
    }
  if(!nameFound)
    {
    vtkstd::cout << "Mandatory element Name not found" << vtkstd::endl;
    return 0;
    }

  return 1;
}

//------------------------------------------------------------------------------
vtkPolyData *xmlCFD::ParseGeometryData(vtkXMLDataElement *geometryData,
bool &containFieldData)
{
  // containFieldData: bool, default: false
  containFieldData = false;
  const char *containFieldDataAttribute
      = geometryData->GetAttribute("containFieldData");
  if(containFieldDataAttribute
      && (vtkStdString(containFieldDataAttribute) == "true"))
    {
    containFieldData = true;
    }

  // geometryFormat: mandatory
  const char *geometryFormat = geometryData->GetAttribute("geometryFormat");
  if(!geometryFormat)
    {
    vtkstd::cout << "Attribute geometryFormat not found" << vtkstd::endl;
    return 0;
    }

  vtkPolyData *data = 0;

  // ExternalSource: optional, only once, takes precedence over Source
  if(vtkXMLDataElement *externalFileName
      = geometryData->FindNestedElementWithName("ExternalSource"))
    {
    if(vtkStdString(geometryFormat) == "vtp")
      {
      vtkXMLPolyDataReader *reader = vtkXMLPolyDataReader::New();

      vtkstd::cout << "GeometryData file name: "
                   << externalFileName->GetCharacterData() << vtkstd::endl;
      reader->SetFileName((*this->Path + "/"
              + externalFileName->GetCharacterData()).c_str());
      if(!reader->GetExecutive()->Update())
        {
        vtkstd::cout << "Error in reading "
                     << externalFileName->GetCharacterData() << vtkstd::endl;
        reader->Delete();
        return 0;
        }
      data = reader->GetOutput();
      if(!data)
        {
        vtkstd::cout << "NULL returned for geometry data "
                     << reader->GetFileName() << vtkstd::endl;
        reader->Delete();
        return 0;
        }
      data->Register(0);
      reader->Delete();
      }
    else
      {
      vtkstd::cout << "Unsupported file format " << geometryFormat
                 << vtkstd::endl;
      return 0;
      }
    }
  else
    {
    // still Source needs to be implemented
    vtkstd::cout << "No ExternalSource found" << vtkstd::endl;
    return 0;
    }

  // transformMatrix: matrix, default: unity
  double mat[12];
  const char *transformMatrix = geometryData->GetAttribute("transformMatrix");
  if(transformMatrix)
    {
    vtksys_ios::istringstream matStream(transformMatrix);
    for(int cmptI = 0; cmptI < 12; ++cmptI)
      {
      matStream >> mat[cmptI];
      }
    vtkFloatArray *pointArray
      = vtkFloatArray::SafeDownCast(data->GetPoints()->GetData());
    if(!pointArray)
      {
      vtkstd::cout << "Data type of the points in geometry is not double"
                   << vtkstd::endl;
      data->Delete();
      return 0;
      }
    float *point = pointArray->GetPointer(0);
    const vtkIdType nPoints = data->GetPoints()->GetNumberOfPoints();
    for(vtkIdType pointI = 0; pointI < nPoints; ++pointI)
      {
      const double x = mat[0] * point[0] + mat[1] * point[1] + mat[2] * point[2]
        + mat[9];
      const double y = mat[3] * point[0] + mat[4] * point[1] + mat[5] * point[2]
        + mat[10];
      const double z = mat[6] * point[0] + mat[7] * point[1] + mat[8] * point[2]
        + mat[11];
      point[0] = x;
      point[1] = y;
      point[2] = z;
      point += 3;
      }
    }

  return data;
}

//------------------------------------------------------------------------------
int xmlCFD::ParseGeometry(vtkXMLDataElement *geometry)
{
  this->ParseName(geometry);

  // GeometryData: mandatory, only once
  if (vtkXMLDataElement *geometryData
      = geometry->FindNestedElementWithName("GeometryData"))
    {
    bool containFieldData; // not used
    this->Boundaries->Geometry
        = this->ParseGeometryData(geometryData, containFieldData);
    return this->Boundaries->Geometry ? 1 : 0;
    }

  return 0;
}

//------------------------------------------------------------------------------
int xmlCFD::ParseModelledGeometry(vtkXMLDataElement *modelledGeometry)
{
  this->ParseName(modelledGeometry);

  // removeInternalCells: optional, default: true
  if (const char *removeInternalCells
      = modelledGeometry->GetAttribute("removeInternalCells"))
    {
    if (vtkStdString(removeInternalCells) != "true")
      {
      vtkstd::cout << "In current version, removeInternalCells must be true"
                   << vtkstd::endl;
      return 0;
      }
    vtkstd::cout << "removeInternalCells: true" << vtkstd::endl;
    }
  else
    {
    vtkstd::cout << "removeInternalCells not found; assuming true"
                 << vtkstd::endl;
    }

  // GeometryData: mandatory, only once
  if (vtkXMLDataElement *geometryData
      = modelledGeometry->FindNestedElementWithName("GeometryData"))
    {
    bool containFieldData; // not used
    this->Boundaries->ModelledGeometry
      = this->ParseGeometryData(geometryData, containFieldData);
    return this->Boundaries->ModelledGeometry ? 1 : 0;
    }

  return 0;
}

//------------------------------------------------------------------------------
int xmlCFD::ParseDimensions(vtkXMLDataElement *field, const char *unit,
    int &dim)
{
  // Dimensions: optional, can appear more than once
  vtkXMLDataElement *dimensions = field->FindNestedElementWithNameAndAttribute(
      "Dimensions", "unit", unit);
  if(!dimensions)
    {
    // Since the element is optional, no problem even if not found
    return 1;
    }

  vtksys_ios::istringstream unitStream(dimensions->GetCharacterData());
  unitStream >> dim;
  if(unitStream.fail())
    {
    vtkstd::cout << "Error in dimension string to number comversion"
                 << vtkstd::endl;
    return 0;
    }

  return 1;
}

//------------------------------------------------------------------------------
int xmlCFD::ParseField(vtkXMLDataElement *field, xmlField &xmlfield)
{
  // fieldIdentifier: mandatory
  const char *fieldIdentifier = field->GetAttribute("fieldIdentifier");
  if(!fieldIdentifier)
    {
    vtkstd::cout << "Attribute fieldIdentifier not found" << vtkstd::endl;
    return 0;
    }
  xmlfield.Identifier = fieldIdentifier;

  // fieldNameInData: optional?
  const char *fieldNameInData = field->GetAttribute("fieldNameInData");
  if(fieldNameInData)
    {
    xmlfield.FieldNameInData = fieldNameInData;
    }
  else
    {
    xmlfield.FieldNameInData = xmlfield.Identifier;
    }

  // fieldType: mandatory, one of scalar, vector, tensor, symTensor
  const char *fieldType = field->GetAttribute("fieldType");
  if(!fieldType)
    {
    vtkstd::cout << "fieldType not found" << vtkstd::endl;
    }
  else if(vtkStdString(fieldType) == "scalar")
    {
    xmlfield.FieldType = xmlField::SCALAR;
    }
  else if(vtkStdString(fieldType) == "vector")
    {
    xmlfield.FieldType = xmlField::VECTOR;
    }
  else if(vtkStdString(fieldType) == "tensor")
    {
    xmlfield.FieldType = xmlField::TENSOR;
    }
  else if(vtkStdString(fieldType) == "symTensor")
    {
    xmlfield.FieldType = xmlField::SYMTENSOR;
    }
  else
    {
    vtkstd::cout << "Unknown field type " << fieldType << vtkstd::endl;
    return 0;
    }

  // useFieldData: optional, default: true
  const char *useFieldData = field->GetAttribute("useFieldData");
  if(useFieldData)
    {
    vtkStdString str(useFieldData);
    if(str == "true")
      {
      xmlfield.UseFieldData = true;
      }
    else if(str == "false")
      {
      xmlfield.UseFieldData = false;
      }
    else
      {
      vtkstd::cout << "Unknown value for useFieldData: " << useFieldData
          << vtkstd::endl;
      return 0;
      }
    }
  else
    {
    xmlfield.UseFieldData = true;
    }

  // scalingFactor: optional, default: 1
  const char *scalingFactor = field->GetAttribute("scalingFactor");
  if(scalingFactor)
    {
    vtksys_ios::istringstream scaleStream(scalingFactor);
    double factor;
    scaleStream >> factor;
    if(scaleStream.fail())
      {
      vtkstd::cout << "Error in reading scaling factor" << vtkstd::endl;
      return 0;
      }
    xmlfield.ScalingFactor = factor;
    }
  else
    {
    xmlfield.ScalingFactor = 1.0;
    }

  // interpolationScheme: optional, default: linear
  const char *interpolationScheme = field->GetAttribute("interpolationScheme");
  if(interpolationScheme)
    {
    xmlfield.InterpolationScheme = interpolationScheme;
    }
  else
    {
    xmlfield.InterpolationScheme = "linear";
    }

  this->ParseName(field);

  static const char *units[] = { "k", "m", "s", "K", "mol", "A", "cd" };

  for(int dimI = 0; dimI < 7; ++dimI)
    {
    if(!this->ParseDimensions(field, units[dimI], xmlfield.Dimensions[dimI]))
      {
      return 0;
      }
    }

  // TODO: SurfaceIntegralValue: optional, only once
  // TODO: VolumeIntegralValue: optional, only once

  // Conditions: mandatory, can appear more than once
  vtkXMLDataElement *conditions
      = field->FindNestedElementWithName("Conditions");
  if(!conditions)
    {
    vtkstd::cout << "Mandatory element Conditions not found" << vtkstd::endl;
    return 0;
    }
  xmlfield.Conditions = conditions->GetCharacterData();

  return 1;
}

//------------------------------------------------------------------------------
int xmlCFD::ParseBoundaryConditions(vtkXMLDataElement *boundaryConditions)
{
  this->Boundaries->push_back(xmlBoundary());
  xmlBoundary &boundary = this->Boundaries->back();

  // boundaryIdentifier: mandatory
  const char *boundaryIdentifier
      = boundaryConditions->GetAttribute("boundaryIdentifier");
  if(!boundaryIdentifier)
    {
    vtkstd::cout << "Attribute boundaryIdentifier not found" << vtkstd::endl;
    return 0;
    }
  vtkstd::cout << "boundary name: " << boundaryIdentifier << vtkstd::endl;
  boundary.Identifier = boundaryIdentifier;

  this->ParseName(boundaryConditions);

  // GeometryData: mandatory, only once
  vtkXMLDataElement *geometryData
      = boundaryConditions->FindNestedElementWithName("GeometryData");
  if (!geometryData)
    {
    vtkstd::cout << "Mandatory element GeometryData not found" << vtkstd::endl;
    return 0;
    }
  bool containFieldData = false;
  boundary.BoundaryGeometry
      = this->ParseGeometryData(geometryData, containFieldData);
  if(boundary.BoundaryGeometry)
    {
    if(containFieldData)
      {
      boundary.FieldData = boundary.BoundaryGeometry;
      boundary.FieldData->Register(0);
      }
    else
      {
      // FieldData: (mandatory), only once
      vtkXMLDataElement *fieldData
          = boundaryConditions->FindNestedElementWithName("FieldData");
      if (fieldData)
        {
        boundary.FieldData
            = this->ParseGeometryData(fieldData, containFieldData);
        if(!boundary.FieldData)
          {
          return 0;
          }
        }
      else
        {
        vtkstd::cout << "Element FildData not found" << vtkstd::endl;
        return 0;
        }
      }
    }
  else
    {
    return 0;
    }

  // Field: mandatory, can appear more than once
  const int nNestedElements = boundaryConditions->GetNumberOfNestedElements();
  bool fieldFound = false;
  for(int elemI = 0; elemI < nNestedElements; ++elemI)
    {
    vtkXMLDataElement *elem = boundaryConditions->GetNestedElement(elemI);
    if(vtkStdString(elem->GetName()) == "Field")
      {
      fieldFound = true;
      boundary.push_back(xmlField());
      xmlField &xmlfield = boundary.back();
      if(!this->ParseField(elem, xmlfield))
        {
        return 0;
        }
      }
    }
  if(!fieldFound)
    {
    vtkstd::cout << "Mandatory element Field not found" << vtkstd::endl;
    return 0;
    }
  
  return 1;
}

//------------------------------------------------------------------------------
int xmlCFD::ParseConditions(vtkXMLDataElement *conditions)
{
  // BoundaryConditions: optional, can appear more than once
  const int nNestedElements = conditions->GetNumberOfNestedElements();
  for(int elemI = 0; elemI < nNestedElements; ++elemI)
    {
    vtkXMLDataElement *elem = conditions->GetNestedElement(elemI);
    if(vtkStdString(elem->GetName()) == "BoundaryConditions")
      {
      if(!this->ParseBoundaryConditions(elem))
        {
        return 0;
        }
      }
    }

  return 1;
}

//------------------------------------------------------------------------------
int xmlCFD::ParseSimulationInputData(vtkXMLDataElement *simulationInputData)
{
  // controlledByCommunication: optional, default = false
  if (const char *controlledByCommunication
      = simulationInputData->GetAttribute("controlledByCommunication"))
    {
    vtkstd::cout << "Controlled by communication: "
                 << controlledByCommunication << vtkstd::endl;
    if (vtkStdString(controlledByCommunication) != "false")
      {
      vtkstd::cout << "Controlled by communication must be false"
                   << vtkstd::endl;
      return 0;
      }
    }

  // Geometry: optional, only once
  if (vtkXMLDataElement *geometry
      = simulationInputData->FindNestedElementWithName("Geometry"))
    {
    if(!this->ParseGeometry(geometry))
      {
      return 0;
      }
    }

  // ModelledGeometry: mandatory, only once
  if (vtkXMLDataElement *modelledGeometry
      = simulationInputData->FindNestedElementWithName("ModelledGeometry"))
    {
    if(!this->ParseModelledGeometry(modelledGeometry))
      {
      return 0;
      }
    }
  else
    {
    vtkstd::cout << "Element ModelledGeometry not found" << vtkstd::endl;
    return 0;
    }

  // SimulationConditions: mandatory, only once
  if (vtkXMLDataElement *simulationConditions
      = simulationInputData->FindNestedElementWithName("SimulationConditions"))
    {
    return this->ParseConditions(simulationConditions);
    }
  else
    {
    vtkstd::cout << "Element SimulationConditions not found" << vtkstd::endl;
    return 0;
    }
}

//------------------------------------------------------------------------------
int xmlCFD::ParseXYZFile(vtkXMLDataElement *xyzFile)
{
  this->ParseName(xyzFile);

  if (vtkXMLDataElement *simulationInputData
      = xyzFile->FindNestedElementWithName("SimulationInputData"))
    {
    return this->ParseSimulationInputData(simulationInputData);
    }

  return 0;
}

//------------------------------------------------------------------------------
int xmlCFD::Parse()
{
  // default encoding set to UTF-8
  if(!this->Superclass::Parse())
    {
    vtkstd::cout << "Error in parsing file " << this->GetFileName()
                 << vtkstd::endl;
    return 0;
    }

  if(vtkXMLDataElement *xyzFile = this->GetRootElement())
    {
    return this->ParseXYZFile(xyzFile);
    }

  return 0;
}
